import socket
import select
from typing import Optional, Callable

from .exceptions import MllpConnectionError


START_BYTE = b'\x0B'
END_BYTES = b'\x1C\x0D'
BUFSIZE = 4096
MAX_MESSAGE_SIZE = 1 * 1024 * 1024  # 1 MB is probably reasonable.


class MllpClient:
    """
    MllpClient provides a simple API for a client to talk to an server, whether an
    integration engine, a RIS, a HIS, a PACS, whatever it is.

    The main assumption made is that messages will be sent one at a time, over a
    single connection. Encryption is not currently supported.

    There is a 1MB limit for the messages out of the box to control the memory usage,
    this can be changed by setting `hl7lw.mllp.MAX_MESSAGE_SIZE` to another value.

    The basic usage goes like:

    ```
    c = MllpClient()
    c.connect(host="127.0.0.1", port=1234)
    c.send(message)
    ack = c.recv()
    c.close()
    ```

    """
    def __init__(self) -> None:
        self.socket: Optional[socket.socket] = None
        self.connected: bool = False
        self.host: Optional[str] = None
        self.port: Optional[int] = None
        self.buffer: bytes = b''

    def is_connected(self) -> bool:
        """
        Used to check if connected.
        """
        return self.connected

    def close(self) -> None:
        """
        Close the connection, also resets the internal state.

        Raises an `MllpConnectionError` if called on a closed connection.
        """
        if self.connected:
            self.connected = False
            self.buffer = b''
            self.socket.close()
        else:
            raise MllpConnectionError("Not connected!")

    def connect(self, host: str, port: int) -> None:
        """
        Connect to an `host` and `port`. Host will be resolved and it resolves to
        multiple IPv4 or IPv6 addresses, they will all be tried before giving up,
        see `socked.create_connection()` for details.

        If already connected, the previous connection is closed and the new connection
        is then opened.

        Network related exceptions will get wrapped into an `MllpConnectionError` and
        then returned.
        """
        self.host = host
        self.port = port
        if self.connected:
            # If we were connected, reset the state.
            self.connected = False
            self.buffer = b''
            self.socket.close()
        try:
            self.socket = socket.create_connection((host, port))
        except TimeoutError as e:
            raise MllpConnectionError(f"Timed out trying to connect to {host}:{port}") from e
        except OSError as e:
            raise MllpConnectionError(f"Failed to connect to {host}:{port}") from e
        self.connected = True

    def send(self, message: bytes, auto_reconnect: bool = True) -> None:
        """
        Send a `message` over the connection.
        
        If not connected and the `auto_reconnect` option is enabled, `connect()` will be
        called with the last connection details used. If `connect()` had yet to be called,
        no connection details will be available and an `MllpConnectionError` will be
        raised instead.

        Should sending fail with an exception, the connection will be closed before the
        exception is re-raised as an `MllpConnectionError`. There's no way to know how
        much of the message was sent.

        MLLP framing is handled by the method.
        """
        if not self.connected:
            if auto_reconnect:
                if self.host is None or self.port is None:
                    raise MllpConnectionError("No host configured!")
                self.connect(host=self.host, port=self.port)
            else:
                raise MllpConnectionError("Not connected!")
        try:
            self.socket.sendall(START_BYTE + message + END_BYTES)
        except Exception as e:
            self.socket.close()
            self.connected = False
            self.buffer = b''
            raise MllpConnectionError("Failed to send message to client.") from e
    
    def recv(self) -> bytes:
        """
        Receive a message from the connection. If there's multiple messages to
        be read, this method needs to be called repeatedly. The method will block
        until a message is available.

        If the connection is not connected, an `MllpConnectionError` will be raised.

        If there's any network error, an `MllpConnectionError` will be raised.

        If a message being read exceeds the maximum size, an `MllpConnectionError`
        will be raised.

        Whenever an exception is raised, the connection will also be closed if it
        wasn't closed already.

        WARNING: You are responsible to use this method to consume the ACKs from
        the other system. If you do not consume the ACKs, eventually the network
        buffers will fill and the other side will block trying to send a ACK. If
        you write an application that loops sending messages and it keeps stalling
        it might be that you are not consumming the ACKs! 
        """
        if not self.connected:
            # No point in connecting. Clients aren't normally polling in MLLP.
            # Maybe if asynch ACKs are used? But this client implementation really
            # isn't that smart.
            raise MllpConnectionError("Not connected!")
        # self.buffer is for any excess bytes after last message. A busy sender that
        # does not expect ack can send messages fast enough they run into each other.
        buffer = self.buffer
        self.buffer = b''
        while True:
            try:
                # This is slow for very large messages.
                buffer += self.socket.recv(BUFSIZE)
            except Exception as e:
                self.connected = False
                self.socket.close()
                raise MllpConnectionError("Failed to read from socket, closing it.") from e
            start = buffer.find(START_BYTE)
            if start == -1:
                # This only happens if buffer is only junk, so get rid of it to minimize
                # memory footprint.
                buffer = b''
            elif start > 0:
                # START_BYTE at index 0 is expected normal when message is split
                buffer = buffer[start:]
            end = buffer.find(END_BYTES)
            if end != -1:
                message = buffer[:end]
                self.buffer = buffer[end:]
                return message[1:]  # Discard leading START_BYTE
            if len(buffer) > MAX_MESSAGE_SIZE:
                self.connected = False
                self.socket.close()
                raise MllpConnectionError(f"Maximum messages size {MAX_MESSAGE_SIZE} exceeded!")


class MllpServer:
    """
    Simple server class to listen for HL7 messages on the port `port` and call the
    callback `callback` on every messages. The callback may return `None` or a `bytes`
    object. If `bytes` are returned, an ack will be sent to the client.

    A trivial callback to create a message sink would look like:

    ```
    from hl7lw.utils import generate_ack, Acks
    from hl7lw import Hl7Parser
    from typing import Optional

    def callback(message: bytes) -> Optional[bytes]:
        p = Hl7Parser()
        m = p.parse_message(message)
        a = generate_ack(m, Acks.AA)
        return p.format_message(a, encoding="ascii")
    ```

    NOTE: The callback is responsible to handle all Exceptions it encounters. Any
    exception that is raised by the callback or not handled by the callback will be
    allowed to bubble up to the caller of `server_forever()` and as such, will kill
    the server.

    It is intentional that MllpServer does not do the HL7 parsing as MLLP can be used
    as transport for non-HL7 messages.
    """
    def __init__(self, port: int, callback: Callable[[bytes], Optional[bytes]]) -> None:
        """
        Initialize the server configuration, providing both the `port` and the `callback`.
        """
        self.port = port
        self.callback = callback
    
    def serve_forever(self):
        """
        Main loop of the server.

        When a message is received, the callback will be called. The server is paused while
        the callback processes. There is absolutely no concurrency in play.
        """
        server_sock = socket.create_server(('', self.port))
        server_sock.setblocking(False)
        clients: dict[socket.socket, tuple[str, int]] = {}
        read_buffers: dict[socket.socket, bytes] = {}
        write_buffers: dict[socket.socket, bytes] = {}
        client_socks = [server_sock]
        while True:
            ready_r, ready_w, _ = select.select(client_socks, write_buffers.keys(), [], 0.1)
            
            for sock in ready_w:
                if sock in write_buffers:
                    buf = write_buffers[sock]
                    try:
                        count = sock.send(buf)
                    except OSError:
                        del clients[sock]
                        del read_buffers[sock]
                        del write_buffers[sock]
                        client_socks.remove(sock)
                        sock.shutdown(socket.SHUT_RDWR)
                        sock.close()
                        continue
                    buf = buf[count:]
                    if len(buf) > 0:
                        write_buffers[sock] = buf
                    else:
                        del write_buffers[sock]
            
            for sock in ready_r:
                if sock == server_sock:
                    conn, addr = sock.accept()
                    clients[conn] = addr
                    client_socks.append(conn)
                    read_buffers[conn] = b''
                    conn.setblocking(False)
                elif sock in read_buffers:
                    buf = read_buffers[sock]
                    try:
                        buf += sock.recv(BUFSIZE)
                    except OSError:
                        del clients[sock]
                        del read_buffers[sock]
                        del write_buffers[sock]
                        client_socks.remove(sock)
                        sock.shutdown(socket.SHUT_RDWR)
                        sock.close()
                        continue
                    index = buf.find(START_BYTE)
                    if index == -1:
                        buf = b''
                    else:
                        buf = buf[index:]
                    index = buf.find(END_BYTES)
                    message = None
                    if index != -1:
                        message = buf[1:index]
                        buf = buf[index:]
                    read_buffers[sock] = buf
                    # We got a message! Process it.
                    if message is not None:
                        ack = self.callback(message)
                        if ack is not None:
                            if sock not in write_buffers:
                                write_buffers[sock] = b''
                            write_buffers[sock] += START_BYTE + ack + END_BYTES




from freezegun import freeze_time
import datetime
from src.hl7lw import Hl7Message, Hl7Segment, Hl7Parser, utils


CONSTANT_TIME = datetime.datetime(year=2024, month=7, day=12, hour=15, minute=6, second=3)


@freeze_time(CONSTANT_TIME)
def test_ack(mocker, trivial_a08: bytes, expected_ack: bytes) -> None:
    p = Hl7Parser()
    m = p.parse_message(trivial_a08)
    # Ensure that the ACK is always identical.
    mocker.patch("random.randint", return_value=999999)
    a = p.format_message(utils.generate_ack(m, utils.Acks.AA), encoding="ascii")
    assert a == expected_ack


@freeze_time(CONSTANT_TIME)
def test_ack_with_details(mocker, trivial_a08: bytes, expected_ack: bytes) -> None:
    p = Hl7Parser()
    m = p.parse_message(trivial_a08)
    # Ensure that the ACK is always identical.
    mocker.patch("random.randint", return_value=999999)
    a = p.format_message(utils.generate_ack(m, utils.Acks.AA, details="bad"))
    expected = p.parse_message(expected_ack)
    expected["MSA-3"] = "bad"
    assert a == str(expected)


def test_patient_id():
    p_w_issuer = utils.PatientID(patient_id="123456", issuer="EPIC")
    p_w_o_issuer = utils.PatientID(patient_id="7890283")

    assert str(p_w_issuer) == "123456^^^EPIC"
    assert str(p_w_o_issuer) == "7890283"


def test_patient():
    patient = utils.Patient(
        patient_ids=[
            utils.PatientID(patient_id="1234", issuer="EPIC"),
            utils.PatientID(patient_id="921380912", issuer="EMPI"),
        ],
        name="Smith^John^Q",
        birthdate="20240823",
        sex="M"
    )
    pid = patient.as_segment()
    assert isinstance(pid, Hl7Segment)
    assert pid.name == 'PID'
    assert str(pid) == "PID|||1234^^^EPIC~921380912^^^EMPI||Smith^John^Q||20240823|M"


@freeze_time(CONSTANT_TIME)
def test_orm_builde_empty(mocker, empty_orm: bytes):
    p = Hl7Parser()
    mocker.patch("random.randint", return_value=999999)
    orm = utils.OrmBuilder(
        sending_application="sa",
        sending_facility="sf",
        receiving_application="ra",
        receiving_facility="rf",
        hl7_version="2.4",
        processing_mode=utils.ProcessingMode.Training
    ).build(parser=p)
    assert p.format_message(orm, encoding="ascii") == empty_orm


@freeze_time(CONSTANT_TIME)
def test_orm_builder_full(mocker, full_orm: bytes):
    p = Hl7Parser()
    mocker.patch("random.randint", return_value=999999)
    orm = utils.OrmBuilder(
        sending_application="sa",
        sending_facility="sf",
        receiving_application="ra",
        receiving_facility="rf",
        hl7_version="2.4",
        processing_mode=utils.ProcessingMode.Training
    ).set_patient(
        patient=utils.Patient(
            patient_ids=[utils.PatientID('1234'), utils.PatientID('456', 'EPIC')],
            name="Smith^John^Q",
            birthdate="18911230",
            sex="M"
        )
    ).set_visit(
        visit=utils.Visit(
            patient_class=utils.PatientClass.Emergency,
            patient_location="EMERG^^^Medical Center",
            referring_physician="1234567^Referring^Doctor",
            visit_number="12345",
            visit_indicator=utils.VisitIndicator.VisitLevel
        )
    ).add_order_group(
        order_group=utils.OrderGroup(
            order_control=utils.OrderControl.ChangeOrder,
            placer_order_number="P1234",
            filler_order_number="F1234",
            accession_number="A1234",
            entering_organization="MCH^Medical Center",
            reason_for_exam="Patient swallowed a whole tuna can"
        ).add_procedure(
            procedure=utils.Procedure(
                requested_procedure_id="RP123401",
                order_status="IP",
                procedure_code="8102",
                procedure_description="Lung",
                quantity_timing=utils.QuantityTiming(start_time="202411171456", end_time="202411171611", priority="R"),
                ordering_provider="12345^Ordering^Provider",
                modality_or_service="CR",
                result_status=None
            )
        ).add_procedure(
            procedure=utils.Procedure(
                requested_procedure_id="RP123402",
                order_status="IP",
                procedure_code="8114",
                procedure_description="Stomach",
                quantity_timing=utils.QuantityTiming(start_time="202411171456", end_time="202411171617", priority="R"),
                ordering_provider="12345^Ordering^Provider",
                modality_or_service="CR",
                result_status=None
            )
        )
    ).build(parser=p)
    assert p.format_message(orm, encoding="ascii") == full_orm


from enum import Enum
from typing import Optional, List
import datetime
import random
from .parser import Hl7Message, Hl7Segment, Hl7Parser, Hl7Field


class Acks(Enum):
    """
    Enum for the acknowledgement codes from HL7 table 0008.
    """
    AA = 1
    AR = 2
    AE = 3
    CA = 4
    CE = 5
    CR = 6


def generate_message_time():
    """
    Generate a current timestamp suitable for MSH-7.
    """
    return datetime.datetime.now().strftime("%Y%m%d%H%M%S")


def generate_message_id():
    """
    Generate a message control ID suitable for MSH-10.

    This function is not suitable for very high volume message generation as
    it generates only 1 million possible ID for a given second and with the
    birthday paradox, you would expect a collision after 1178 messages, 50%
    of the time.
    """
    return f"{generate_message_time()}{random.randint(0, 999999):06d}"


def generate_ack(message: Hl7Message, status: Acks, details: Optional[str] = None) -> Hl7Message:
    """
    Generate an acknowledgement for `message` with the acknowledgement code `status`.
    An optional `reason` can be supplied and if so it will be put into MSA-3.

    This is a convenience function for implementing simple Hl7 sinks.
    """
    ack = Hl7Message(parser=message.parser)
    orig_msh = message.get_segment('MSH', strict=False)
    
    msh = Hl7Segment(parser=message.parser)
    msh.parse(str(orig_msh))
    msh[3] = orig_msh[5]
    msh[4] = orig_msh[6]
    msh[5] = orig_msh[3]
    msh[6] = orig_msh[4]
    msh[7] = generate_message_time()
    msh[9] = 'ACK'
    msh[10] = generate_message_id()
    
    msa = Hl7Segment(parser=message.parser)
    msa.parse("MSA|||")
    msa[1] = status.name
    msa[2] = orig_msh[10]
    if details is not None:
        msa[3] = details
    
    ack.segments = [msh, msa]
    return ack


class PatientID:
    def __init__(self, patient_id: str, issuer: Optional[str] = None) -> None:
        self.patient_id = patient_id
        self.issuer = issuer
    
    def __str__(self) -> str:
        f = Hl7Field(Hl7Parser())
        f[1][1] = self.patient_id
        if self.issuer is not None:
            f[1][4] = self.issuer
        return str(f)


class Patient:
    def __init__(self, patient_ids: List[PatientID], name: str, birthdate: str, sex: str) -> None:
        self.patient_ids = patient_ids
        self.name = name
        self.birthdate = birthdate
        self.sex = sex
    
    def as_segment(self, parser: Optional[Hl7Parser] = None) -> Hl7Segment:
        if parser is None:
            parser = Hl7Parser()
        pid = Hl7Segment(parser=parser)
        pid.parse('PID|')
        pid3 = Hl7Field(parser)
        for index, patient_id in enumerate(self.patient_ids, start=1):
            pid3[index] = str(patient_id)
        pid[3] = str(pid3)
        pid[5] = self.name
        pid[7] = self.birthdate
        pid[8] = self.sex
        return pid


class VisitIndicator(Enum):
    AccountLevel = "A"
    VisitLevel = "V"


class PatientClass(Enum):
    Emergency = "E"
    Inpatient = "I"
    Outpatient = "O"
    Preadmit = "P"
    RecurringPatient = "R"
    Obstetrics = "O"
    CommercialAccount = "C"
    NotApplicable = "N"
    Unknown = "U"


class Visit:
    def __init__(self,
                 patient_class: PatientClass = PatientClass.Outpatient,
                 patient_location: str = "",
                 referring_physician: str = "",
                 visit_number: str = "",
                 visit_indicator: VisitIndicator = VisitIndicator.AccountLevel
                 ) -> None:
        self.patient_class = patient_class
        self.patient_location = patient_location
        self.referring_physician = referring_physician
        self.visit_number = visit_number
        self.visit_indicator = visit_indicator
    
    def as_segment(self, parser: Optional[Hl7Parser] = None) -> Hl7Segment:
        if parser is None:
            parser = Hl7Parser()
        pv1 = Hl7Segment(parser=parser)
        pv1.parse('PV1|')
        pv1[2] = self.patient_class.value
        pv1[3] = self.patient_location
        pv1[8] = self.referring_physician
        pv1[19] = self.visit_number
        pv1[51] = self.visit_indicator.value
        return pv1


class OrderControl(Enum):
    """
    From HL7 Table 0119

    For brevity, I kept Order whenevent the table has Order/Service Request
    """
    NewOrder = "NW"
    OrderOK = "OK"
    UnableToAccept = "UA"
    PreviousResult = "PR"
    CancelOrder = "CA"
    OrderCancelled = "OC"
    CancelledAsRequested = "CR"
    UnableToCancel = "UC"
    DiscontinueOrder = "DC"
    OrderDiscontinued = "OD"
    DiscontinuedAsRequested = "DR"
    UnableToDiscontinue = "UD"
    Hold = "HD"
    OrderHeld = "OH"
    UnableToPutOnHold = "UH"
    OnHoldAsRequested = "HR"
    ReleasePreviousHold = "RL"
    OrderReleased = "OE"
    ReleasedAsRequested = "OR"
    UnableToRelease = "UR"
    OrderReplaceRequest = "RP"
    ReplaceUnsolicited = "RU"
    ReplacementOrder = "RO"
    ReplacedAsRequested = "RQ"
    UnableToReplace = "UM"
    ParentOrder = "PA"
    ChildOrder = "CH"
    ChangeOrder = "XO"
    OrderChangedUnsolicited = "XX"
    UnableToChange = "UX"
    ChangedAsRequested = "XR"
    DataErrors = "DE"
    ObservationToFollow = "RE"
    RequestReceived = "RR"
    ResponseToSendStatusRequest = "SR"
    SendStatusRequest = "SS"
    StatusChanged = "SC"
    SendOrderNumber = "SN"
    NumberAssigned = "NA"
    CombinedResult = "CN"
    RefillOrder = "RF"
    OrderRefillRequestApproval = "AF"
    OrderRefillRequestDenied = "DF"
    OrderRefilledUnsolicited = "FU"
    OrderRefilledAsRequested = "OF"
    UnableToRefill = "UF"
    LinkOrderToPatientCareGoal = "LI"
    UnlinkOrderFromPatientCareGoal = "UN"


class ResultStatus(Enum):
    """
    HL7 Table 0123
    """
    OrderReceived = "O"
    ProcedureIncomplete = "I"
    ProcedureScheduled = "S"
    SomeResultsAvailable = "A"
    Preliminary = "P"
    CorrectionToResults = "C"
    ResultsStoredNotYetVerified = "R"
    Final = "F"
    OrderCancelled = "X"


class QuantityTiming:
    def __init__(self, start_time: str = "", end_time: str = "", priority: str = "") -> None:
        self.start_time = start_time
        self.end_time = end_time
        self.priority = priority
        
    def __str__(self) -> str:
        f = Hl7Field(Hl7Parser())
        f[1][4] = self.start_time
        f[1][5] = self.end_time
        f[1][6] = self.priority
        return str(f)


class Procedure:
    def __init__(self,
                 requested_procedure_id: str = "",
                 order_status: str = "",  # HL7 Table 0038 is ignored too often to codify
                 procedure_code: str = "",
                 procedure_description: str = "", 
                 quantity_timing: Optional[QuantityTiming] = None,
                 ordering_provider: str = "",
                 modality_or_service: str = "",
                 result_status: Optional[ResultStatus] = None
                 ) -> None:
        self.requested_procedure_id = requested_procedure_id
        self.order_status = order_status
        self.procedure_code = procedure_code
        self.procedure_description = procedure_description
        self.quantity_timing = quantity_timing
        self.ordering_provider = ordering_provider
        self.modality_or_service = modality_or_service
        self.result_status = result_status
    
    def as_universal_service_id(self):
        f = Hl7Field(Hl7Parser())
        f[1][1] = self.procedure_code
        f[1][2] = self.procedure_description
        return str(f)


class OrderGroup:
    def __init__(self,
                 order_control: OrderControl = OrderControl.NewOrder,
                 placer_order_number: str = "",
                 filler_order_number: str = "",
                 accession_number: str = "",
                 entering_organization: str = "",
                 reason_for_exam: str = "",
                 ) -> None:
        self.procedures: list[Procedure] = []
        self.order_control = order_control
        self.placer_order_number = placer_order_number
        self.filler_order_number = filler_order_number
        self.accession_number = accession_number
        self.entering_organization = entering_organization
        self.reason_for_exam = reason_for_exam

    def add_procedure(self, procedure: Procedure) -> "OrderGroup":
        self.procedures.append(procedure)
        return self

    def add_to_message(self, message: Hl7Message) -> None:
        base_orc = message.parser.parse_segment('ORC|')
        base_obr = message.parser.parse_segment('OBR|')
        
        base_orc[1] = self.order_control.value
        base_orc[2] = self.placer_order_number
        base_orc[3] = self.filler_order_number
        base_orc[17] = self.entering_organization

        base_obr[1] = "1"
        base_obr[2] = self.placer_order_number
        base_obr[3] = self.filler_order_number
        base_obr[18] = self.accession_number
        reason = Hl7Field(parser=message.parser)
        reason[1][2] = self.reason_for_exam
        base_obr[31] = str(reason)

        if len(self.procedures) == 0:
            message.segments.append(base_orc)
            message.segments.append(base_obr)
        else:
            segments_to_add = []
            p = message.parser
            for index, procedure in enumerate(self.procedures, start=1):
                # Reparse to make deep copies.
                orc = p.parse_segment(p.format_segment(base_orc))
                obr = p.parse_segment(p.format_segment(base_obr))

                if procedure.quantity_timing is not None:
                    orc[7] = str(procedure.quantity_timing)
                    obr[27] = orc[7]
                    obr[5] = procedure.quantity_timing.priority
                    obr[6] = procedure.quantity_timing.start_time
                    obr[7] = procedure.quantity_timing.start_time
                    obr[8] = procedure.quantity_timing.end_time
                    orc[15] = procedure.quantity_timing.start_time
                    obr[36] = procedure.quantity_timing.start_time
                if procedure.result_status is not None:
                    obr[25] = procedure.result_status.value
                obr[4] = procedure.as_universal_service_id()
                orc[5] = procedure.order_status
                orc[12] = procedure.ordering_provider
                obr[16] = procedure.ordering_provider
                obr[19] = procedure.requested_procedure_id
                obr[24] = procedure.modality_or_service
                obr[44] = procedure.procedure_code

                segments_to_add.append(orc)
                segments_to_add.append(obr)

            message.segments += segments_to_add


BASE_MSH = b'MSH|^~\\&|||||||||T|2.3.1||||||||||\r'


class ProcessingMode(Enum):
    Debugging = 'D'
    Production = 'P'
    Training = 'T'


class OrmBuilder:
    def __init__(self,
                 sending_application: str = "hl7lw",
                 sending_facility: str = "",
                 receiving_application: str = "",
                 receiving_facility: str = "",
                 hl7_version: str = "2.3.1",
                 processing_mode: ProcessingMode = ProcessingMode.Production 
                 ) -> None:
        self.sending_application = sending_application
        self.sending_facility = sending_facility
        self.receiving_application = receiving_application
        self.receiving_facility = receiving_facility
        self.hl7_version = hl7_version
        self.processing_mode = processing_mode

        self.patient: Optional[Patient] = None
        self.visit: Optional[Visit] = None
        self.order_group: Optional[OrderGroup] = None
    
    def set_patient(self, patient: Patient) -> "OrmBuilder":
        self.patient = patient
        return self
        
    def set_visit(self, visit: Visit) -> "OrmBuilder":
        self.visit = visit
        return self

    def add_order_group(self, order_group: OrderGroup) -> "OrmBuilder":
        self.order_group = order_group
        return self
    
    def build(self, parser: Optional[Hl7Parser] = None) -> Hl7Message:
        if parser is None:
            parser = Hl7Parser()
        m = parser.parse_message(BASE_MSH)
        msh = m.get_segment('MSH')
        msh[3] = self.sending_application
        msh[4] = self.sending_facility
        msh[5] = self.receiving_application
        msh[6] = self.receiving_facility
        msh[7] = generate_message_time()
        msh[9] = 'ORM^O01'
        msh[10] = generate_message_id()
        msh[11] = self.processing_mode.value
        msh[12] = self.hl7_version

        if self.patient is None:
            m.segments.append(parser.parse_segment("PID|"))
        else:
            m.segments.append(self.patient.as_segment(parser=parser))

        if self.visit is not None:
            m.segments.append(self.visit.as_segment(parser=parser))

        if self.order_group is None:
            m.segments.append(parser.parse_segment("ORC|"))
            m.segments.append(parser.parse_segment("OBR|"))
        else:
            self.order_group.add_to_message(message=m)
        
        return m
